"""Defines a base built-in environment, with all the primitives."""


import types
import math
import sys

import pogo
from symbol import Symbol, true, false
import expressions
import environment
import pair
from parser import parse
import evaluator
from error import SchemeError
import operator



## TODO: add eval and apply as primitive procedures

def allNumbers(numbers):
    for n in numbers:
        if type(n) not in (types.IntType, types.LongType, types.FloatType):
            return 0
    return 1


def schemeAdd(*numbers):
    if len(numbers) == 0: return 0
    if not allNumbers(numbers):
        raise SchemeError, "primitive '+' --- arguments must be numbers"
    return reduce(operator.add, numbers)

def schemeSubtract(*numbers):
    if len(numbers) == 0:
        raise SchemeError, "primitive '-' --- expects 1 or more arguments"
    if not allNumbers(numbers):
        raise SchemeError, "primitive '-' --- arguments must be numbers"
    return reduce(operator.sub, numbers)

def schemeMultiply(*numbers):
    if len(numbers) == 0: return 1
    if not allNumbers(numbers):
        raise SchemeError, "primitive '*' --- arguments must be numbers"
    return reduce(operator.mul, numbers)

def schemeDivide(*numbers):
    if len(numbers) == 0:
        raise SchemeError, "primitive '/' --- expects 1 or more arguments"
    if not allNumbers(numbers):
        raise SchemeError, "primitive '/' --- arguments must be numbers"
    return reduce(operator.div, numbers)

def schemeRemainder(a, b):
    if not allNumbers([a, b]):
        raise SchemeError, "primitive 'remainder' --- arguments must be numbers"
    return a % b

def schemeQuotient(a, b):
    if not allNumbers([a, b]):
        raise SchemeError, "primitive 'quotient' --- arguments must be numbers"
    return int(a / b)

def schemeSqrt(a):
    if not allNumbers([a]):
        raise SchemeError, "primitive 'sqrt' --- argument must be a number"
    return math.sqrt(a)

def schemeFloor(a):
    if not allNumbers([a]):
        raise SchemeError, "primitive 'floor' --- argument must be a number"
    return int(math.floor(a))

def schemeCeiling(a):
    if not allNumbers([a]):
        raise SchemeError, "primitive 'ceiling' --- argument must be a number"
    return int(math.ceil(a))

def schemeNumericalEq(a, b):
    """Here, we need to make sure that we're returning the metacircular
    values of true or false."""
    return schemeBooleanize(a == b)

def schemeLessThan(a, b): return schemeBooleanize(a < b)
def schemeLessThanOrEquals(a, b): return schemeBooleanize(a <= b)
def schemeGreaterThan(a, b): return schemeBooleanize(a > b)
def schemeGreaterThanOrEquals(a, b): return schemeBooleanize(a >= b)

def schemeEqQuestion(a, b):
    if type(a) in (types.IntType, types.FloatType, types.LongType):
        return schemeBooleanize(a == b)
    return schemeBooleanize(a is b)

def schemeEqualQuestion(a, b): return schemeBooleanize(a == b)
def schemeListQuestion(a): return schemeBooleanize(pair.isList(a))
def schemeNullQuestion(a): return schemeBooleanize(pair.isNull(a))


def schemeDisplay(thing):
    sys.stdout.write(expressions.expressionToString(thing,
                                                    quoteStrings=0))
    return Symbol("ok")


def schemeWrite(thing):
    sys.stdout.write(expressions.expressionToString(thing,
                                                    quoteStrings=1))
    return Symbol("ok")


def schemeNewline():
    sys.stdout.write("\n")
    return Symbol("ok")

def schemeStringToSymbol(mystr): return Symbol(mystr)
def schemeSymbolToString(mysymbol): return str(mysymbol)
def schemeNumberToString(mynumber): return str(mynumber)

def schemeStringAppend(*strings):
    if len(strings) == 0: return ""
    return reduce(operator.add, strings)


def schemeNot(x):
    if x == false: return true
    return false

def schemeQuit():
    sys.stdout.write("bye\n")
    sys.exit(0)

def schemeBooleanize(x):
    """Translates what Python considers true and false to the Scheme
    symbols "true" or "false"."""
    if x: return true
    return false


def schemePairQuestion(L):
    return schemeBooleanize(pair.isPair(L))


def schemeError(*args):
    raise SchemeError, ' '.join(map(expressions.expressionToString, args))


def wrapPrimitiveWithDefaults(f):
    """Normal primitives don't know what to do with their continuation
    or environment arguments, so this wrapper provides a nice
    default."""
    def newPrimitive(cont, env, args):
        return pogo.apply_cont(cont, (f(*args)))
    return newPrimitive



def schemeDir(cont, env, args):
    """A quick function to inspect the first frame."""
    if len(args) != 0:
        raise TypeError, ("dir expected at most 0 arguments, got %s" % len(args))
    names = environment.firstFrame(env).keys()
    names.sort()
    return pogo.apply_cont(cont, pair.schemeList(* map(Symbol, names)))


def schemeLoad(cont, env, args):
    """Special primitive: implements LOAD."""
    symbolicFilename = str(args[0])
    try:
        f = open(symbolicFilename)
        try:
            text = "(begin \n%s\n 'ok)" % f.read()
        finally:
            f.close()
        return evaluator.teval(parse(text), env, cont)
    except IOError, e:
        raise SchemeError, "LOAD error -- %s" % str(e)
    except TypeError, e:
        raise SchemeError, "LOAD error -- argument must be a string"



def schemeCallcc(cont, env, args):
    lambdaBody = args[0]
    def c2(val):
        return pogo.apply_cont(cont, val)
    cont_procedure = expressions.makeContinuationProcedure(c2)
    return evaluator.apply(lambdaBody, pair.schemeList(cont_procedure),
                           env, cont)


######################################################################




def setupEnvironment():
    """Sets up a new environment with a bunch of fairly standard
    Scheme built-in primitives."""
    PRIMITIVE_PROCEDURES = [
        [Symbol("car"), pair.schemeCar],
        [Symbol("cdr"), pair.schemeCdr],
        [Symbol("cons"), pair.schemeCons],
        [Symbol("append"), pair.schemeAppend],
        [Symbol("list"), pair.schemeList],
        [Symbol("set-car!"), pair.schemeSetCarBang],
        [Symbol("set-cdr!"), pair.schemeSetCdrBang],
        [Symbol("+"), schemeAdd],
        [Symbol("-"), schemeSubtract],
        [Symbol("*"), schemeMultiply],
        [Symbol("/"), schemeDivide],
        [Symbol("remainder"), schemeRemainder],
        [Symbol("quotient"), schemeQuotient],
        [Symbol("sqrt"), schemeSqrt],
        [Symbol("floor"), schemeFloor],
        [Symbol("ceiling"), schemeCeiling],
        [Symbol("="), schemeNumericalEq],
        [Symbol("<"), schemeLessThan],
        [Symbol("<="), schemeLessThanOrEquals],
        [Symbol(">"), schemeGreaterThan],
        [Symbol(">="), schemeGreaterThanOrEquals],
        [Symbol("eq?"), schemeEqQuestion],
        [Symbol("equal?"), schemeEqualQuestion],
        [Symbol("list?"), schemeListQuestion],
        [Symbol("pair?"), schemePairQuestion],
        [Symbol("null?"), schemeNullQuestion],
        [Symbol("display"), schemeDisplay],
        [Symbol("write"), schemeWrite],
        [Symbol("newline"), schemeNewline],
        [Symbol("not"), schemeNot],
        [Symbol("string->symbol"), schemeStringToSymbol],
        [Symbol("symbol->string"), schemeSymbolToString],
        [Symbol("number->string"), schemeNumberToString],
        [Symbol("string-append"), schemeStringAppend],
        [Symbol("quit"), schemeQuit],
        [Symbol("exit"), schemeQuit],
        [Symbol("error"), schemeError],
        ]
    PRIMITIVE_PROCEDURE_NAMES = map(lambda x: x[0], PRIMITIVE_PROCEDURES)
    PRIMITIVE_PROCEDURE_OBJECTS = []
    for name, proc in PRIMITIVE_PROCEDURES:
        schemePrimitiveProcedure = expressions.makePrimitiveProcedure(wrapPrimitiveWithDefaults(proc))
        PRIMITIVE_PROCEDURE_OBJECTS.append(schemePrimitiveProcedure)

    ## Adds some unusual cases here.
    ## LOAD is weird because it does use the environment
    PRIMITIVE_PROCEDURE_NAMES.append(Symbol("load"))
    PRIMITIVE_PROCEDURE_OBJECTS.append(expressions.makePrimitiveProcedure(schemeLoad))
    ## and CALL/CC is just weird, period.  *grin*
    PRIMITIVE_PROCEDURE_NAMES.append(Symbol("call/cc"))
    PRIMITIVE_PROCEDURE_OBJECTS.append(expressions.makePrimitiveProcedure(schemeCallcc))
    PRIMITIVE_PROCEDURE_NAMES.append(Symbol("call-with-current-continuation"))
    PRIMITIVE_PROCEDURE_OBJECTS.append(expressions.makePrimitiveProcedure(schemeCallcc))
    ## dir is just something I wanted to play with.
    PRIMITIVE_PROCEDURE_NAMES.append(Symbol("dir"))
    PRIMITIVE_PROCEDURE_OBJECTS.append(expressions.makePrimitiveProcedure(schemeDir))


    initial_environment = environment.extendEnvironment(
        PRIMITIVE_PROCEDURE_NAMES,
        PRIMITIVE_PROCEDURE_OBJECTS,
        environment.THE_EMPTY_ENVIRONMENT)
    environment.defineVariable(Symbol("#t"), Symbol("#t"), initial_environment)
    environment.defineVariable(Symbol("#f"), Symbol("#f"), initial_environment)
    return initial_environment
