"""Implementation of pairs for Python.

FIXME: not all the primitives are tail recursive!  Must fix this!
"""

from sets import Set

from symbol import Symbol
from error import SchemeError
import pogo
import unittest


## The NIL atom
NIL = []


class Cons(list): pass


def isNull(x):
    return x is NIL


def isPair(p):
    return type(p) is Cons


def schemeCons(head, rest):
    """Returns the concatentation of head with rest."""
    return Cons([head, rest])



def schemeCar(p):
    if not isPair(p):
        raise SchemeError, "CAR --- argument is not a pair."
    if p is NIL:
        raise SchemeError, "CAR --- cannot take CAR of empty list."
    else:
        return p[0]


def schemeCdr(p):
    if not isPair(p):
        raise SchemeError, "CDR --- argument is not a pair."
    if p is NIL:
        raise SchemeError, "CDR --- cannot take CDR of empty list."
    else:
        return p[1]


def schemeCddr(p):
    return schemeCdr(schemeCdr(p))

def schemeCdddr(p):
    return schemeCdr(schemeCdr(schemeCdr(p)))


def schemeCadr(p):
    return schemeCar(schemeCdr(p))

def schemeCaddr(p):
    return schemeCar(schemeCdr(schemeCdr(p)))

def schemeCadddr(p):
    return schemeCar(schemeCdr(schemeCdr(schemeCdr(p))))




def schemeSetCarBang(pair, element):
    if not isPair(pair):
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[0] = element
    return Symbol("ok")


def schemeSetCdrBang(pair, element):
    if not isPair(pair):
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[1] = element
    return Symbol("ok")



## Hmm... some of these functions really belong in builtins.py.

def isList(p):
    """Returns True if p refers to a list-like structure.
    Note: loopy structures don't qualify as lists.
    """
    seenPairIds = Set()
    while True:
        if id(p) in seenPairIds:
            return 0
        seenPairIds.add(id(p))
        if isNull(p): return 1
        if not isPair(p): return 0
        p = schemeCdr(p)


def isDottedPair(p):
    """Returns True if p refers to an improper list."""
    if not isPair(p): return 0
    elif isPair(schemeCdr(p)) or isNull(schemeCdr(p)): return 0
    else: return 1


def schemeLength(p):
    """Returns the length of p.  Assumes that p is a list."""
    if not isList(p): raise SchemeError, "LENGTH --- not a list"
    length = 0
    while True:
        if isNull(p): return length
        length += 1
        p = schemeCdr(p)


def schemeReverse(p):
    """Reverses a list."""
    if not isList(p): raise SchemeError, "REVERSE --- not a list"
    result = NIL
    while not isNull(p):
        result = schemeCons(schemeCar(p), result)
        p = schemeCdr(p)
    return result



def listMap(f, p):
    """Maps a function f across p."""
    if not isList(p): raise SchemeError, "MAP --- not a list"
    resultsRev = NIL
    while not isNull(p):
        resultsRev = schemeCons(f(schemeCar(p)), resultsRev)
        p = schemeCdr(p)
    return schemeReverse(resultsRev)



def c_listMap(c_f, p, cont):
    """Maps a function f across p, but in a continuation-passed style.

    'c_f' is a function that takes an element and the continuation.
    """
    if not isList(p): raise SchemeError, "CMAP --- not a list"
    if isNull(p):
        return pogo.apply_cont(cont, NIL)
    def c_head(head_val):
        def c_tail(tail_val):
            return pogo.apply_cont(cont, schemeCons(head_val, tail_val))
        return pogo.apply_cont(c_listMap, c_f, schemeCdr(p), c_tail)
    return pogo.apply_cont(c_f, schemeCar(p), c_head)



def schemeCount(x, p):
    """Returns the number of elements in p that are equal to x."""
    if not isList(p): raise SchemeError, "COUNT --- not a list"
    result = 0
    while not isNull(p):
        if x == schemeCar(p):
            result += 1
        p = schemeCdr(p)
    return result
    

    
def schemeAppend(*lists):
    """Appends all lists together."""
    appended_list = lists[-1]
    for i in xrange(len(lists)-2, -1, -1):
        next_list = lists[i]
        appended_list = pogo.pogo_start(c_appendTwo,
                                        next_list, appended_list,
                                        pogo.land)
    return appended_list


def c_appendTwo(front, back, cont):
    """Appends two lists together.  Written in continuation passing style."""
    if not isList(front): raise SchemeError, "MAP --- not a list"
    if not isList(back): raise SchemeError, "MAP --- not a list"
    if isNull(front):
        return pogo.apply_cont(cont, back)
    def c(appendedRest):
        return pogo.apply_cont(cont, schemeCons(schemeCar(front),
                                                appendedRest))
    return pogo.apply_cont(c_appendTwo, schemeCdr(front), back, c)



def pairsToPythonList(pair):
    """Does a shallow conversion of a pair list chain to a Python list."""
    if not isList(pair): raise SchemeError, "not a list"
    elements = []
    while not isNull(pair):
        elements.append(schemeCar(pair))
        pair = schemeCdr(pair)
    return elements


def schemeList(*elements):
    """Does a shallow conversion of a Python list to a pair chain."""
    result = list(elements)
    result.reverse()
    return reduce(lambda x, y: schemeCons(y, x), [NIL] + result)


class PairTests(unittest.TestCase):
    def testNull(self):
        self.assert_(isList(NIL))

    def testReversal(self):
        self.assertEquals(schemeList(1, 2, 3, 4, 5),
                          schemeReverse(schemeList(5, 4, 3, 2, 1)))

    def testMapping(self):
        self.assertEquals(schemeList(2, 4, 6, 8),
                          listMap(lambda x: x*2, schemeList(1, 2, 3, 4)))

    def testAppendTwo(self):
        self.assertEquals(schemeList(1, 2, 3, 4),
                          pogo.pogo_start(c_appendTwo,
                                          schemeList(1, 2),
                                          schemeList(3, 4),
                                          pogo.land))


    def testContinuationMapping(self):
        def c_square(x, cont):
            return pogo.apply_cont(cont, x**2)
        self.assertEquals(schemeList(1, 4, 9, 16),
                          pogo.pogo_start(c_listMap,
                                          c_square,
                                          schemeList(1, 2, 3, 4),
                                          pogo.land))


if __name__ == '__main__':
    unittest.main()
