#!/usr/bin/env python

"""A translation of a Scheme interpreter in Python, modeling the
metacircular evaluator in Ch. 4 of The Structure and Interpretation of
Computer Programs.  (http://www-mitpress.mit.edu/sicp/)

AUTHOR:

    Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

SYNOPSIS:

    It's Scheme.  *grin*  At least, quite a bit of Scheme.

    Scheme is a variant of Lisp, and it's used in many introductory CS
    courses.  It is pretty powerful despite its deceptive size.  For
    more information on Scheme, visit the Scheme home page:

        http://www.swiss.ai.mit.edu/projects/scheme


OLD BUGS:

    One major deficiency of this implementation had in the past was
    the lack of tail call elimination.  However, I've gotten this to
    work by using trampolined style.  This involved rewriting the
    interpreter into CPS form, and trampolining the resulting tail
    calls.  The resulting interpreter looks quite a bit uglier, but at
    least it shouldn't stack overflow easily.

BUGS:

    Not all of R5RS is implemented, and probably will never be until I
    finish reading and understanding it.


MISCELLANEOUS:

    This source code is for the public domain; I'm writing it just for
    the fun of it.  This is certainly not meant to be efficient or
    even useful.  *grin*

    My future goals is to make this usable enough to run something
    like the explicit-register simulator and compiler in SICP Ch. 5.
    I'm learning more about the advanced features in scheme
    (continuations, hygenic macros), so I'll probably use this as a
    future testbed to make sure I understand those ideas.
"""




import sys
import optparse
import symbol
import parser
import evaluator
import builtins
import analyzer
import analyzer_builtins
import prompt
import error
import expressions
import pair
######################################################################


class Interpreter:
    """Meant to be subclassed.
    subclasses should define _eval and _env in their constructors."""
    def eval(self, exp):
        return self.getEvaluator()(exp, self.getEnvironment())

    def getEvaluator(self): return self._eval
    def getEnvironment(self): return self._env


class RegularInterpreter(Interpreter):
    """This interpreter evaluates expressions in a way similar to
    the metacircular evaluator described by SICP."""
    def __init__(self):
        self._eval = evaluator.eval
        self._env = builtins.setupEnvironment()


class AnalyzingInterpreter(Interpreter):
    """This version of the interpreter will do syntax analysis on its
    expressions before it executes them.  Should be a win on speed."""
    def __init__(self):
        self._eval = analyzer.eval
        self._env = analyzer_builtins.setupEnvironment()                       



def repl(interp):
    """A fairly standard 'read->eval->print' loop."""
    print 'Welcome to PyScheme!  Type: (QUIT) to quit.\n'
    def promptCallback(s):
        try:
            print expressions.expressionToString(interp.eval(parser.parse(s)))
        except (error.SchemeError, parser.ParserError), e:
            print "ERROR (%s): %s" % (e.__class__, e)

    p = prompt.Prompt(name="PyScheme", callback=promptCallback,
                      quit_str=None)
    p.promptLoop()
            

def isNonInteractive(args):
    """Returns True if we want to run in non-interactive mode."""
    if len(args) >= 1:
        return True


def runNonInteractive(interp, args):
    """Executes the interpreter on args[0]."""
    interp.eval(pair.schemeList(symbol.Symbol("load"),
                                args[0]))


def chooseInterpreter(options):
    """Choose which interpreter we should use."""
    if options.interpreter_type == 'regular':
        return RegularInterpreter()
    else:
        return AnalyzingInterpreter()


def configureOptionParser():
    optparser = optparse.OptionParser()
    optparser.add_option("--regular",
                      help="invoke regular interpreter",
                      action="store_const",
                      const="regular",
                      dest="interpreter_type",
                      )
    optparser.add_option("--analyzer",
                      help="invoke analyzing interpreter [default]",
                      action="store_const",
                      const="analyzer",
                      dest="interpreter_type",
                      )
    return optparser


def main():
    """Main driver."""
    optparser = configureOptionParser()
    options, args = optparser.parse_args()
    interp = chooseInterpreter(options)

    if isNonInteractive(args):
        runNonInteractive(interp, args)
    else:
        repl(interp)



if __name__ == '__main__':
    main()
